/* sched.c
 * simple scheduler main prog
 */
 
#define STDINFILE	"sched.in"
#define STDRDFILE	"sched-rands.txt"
#define STDRR_Q		4

 
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>

#include "proc.h"


int udri(int r, FILE * f);
int sched_next(int algo, int t, int p, pproc ps[], int max);
int min(int a, int b);
void usage();

int main(int argc, char ** argv){

	char * in_file = malloc(32);
	char * rd_file = malloc(32);
	strcpy(in_file,STDINFILE);
	strcpy(rd_file,STDRDFILE);
	
	int rr_q = STDRR_Q;
	int sched_meth = UNI1;
	FILE *infile_fd, *randfile_fd;

	int bufsize = 32;
	char lineptr[bufsize];
	pproc procs[20];
	int numPs = 0,i;

	int index,c;
	//get options
	while ((c = getopt (argc, argv, "a:q:i:r:")) != -1)
		switch (c)
      	{
      		case 'a':			//sched algo
        		sched_meth = atoi(optarg);
        		break;
      		case 'q':			//q for RR
		        rr_q = atoi(optarg);
        		break;
		    case 'i':			//mix infile
        		in_file = optarg;
		        break;
      		case 'r':			//random infile
        		rd_file = optarg;
        		break;
      		case 'h':			//help
        		usage();
        		exit(1);
      		case '?':
        		if (isprint (optopt))
			    	fprintf (stderr, "Unknown option `-%c'.\n", optopt);
        		else
          			fprintf (stderr,
                   		"Unknown option character `\\x%x'.\n",
                   		optopt);
        		usage();
        		exit(1);
      		default:
        		usage();
        		exit(1);
      	}
	for (index = optind; index < argc; index++)
    	printf ("Non-option argument %s\n", argv[index]);
  	
	//open the files
	char opentype = 'r';
	infile_fd = fopen((const char *)in_file, &opentype);
	randfile_fd = fopen((const char *)rd_file, &opentype);
	//check if opened
	if (infile_fd == 0){
		perror("Coulnd't open scheduling file");
		exit(128);
	}
	if (randfile_fd == 0){
		perror("Coulnd't open random num file");
		exit(128);
	}
	
	//now read jobs to schedule
	while ( 0 < fgets((char *)&lineptr,bufsize,infile_fd) ){
		int a,b,c,d,i=0;
		if (sscanf(lineptr,"(%i %i %i %i)",&a,&b,&c,&d) == 4){
			//alloc a new proc obj
			pproc p = (pproc)malloc(sizeof(struct proc));
			//set read vars
			p->A_arriv = a;
			p->B_burst_max = b;
			p->C_total = c;
			p->IO_max = d;
			//and init counters, state etc.
			p->state = READY;
			p->blocked_since = 0;
			p->blocked_till = 0;
			
			p->ready_since = a;
			p->tot_io_time = 0;
			p->tot_ready_time = 0;
			p->tot_running = 0;
			//add to array
			procs[numPs++] = p;
			//fprintf(stdout,"Created p[%i] with a=%i,b=%i,c=%i,d=%i,ptr=%i\n",(numPs-1),a,b,c,d,p);
			
		}else{
			fprintf(stderr,"Error while parsing following line of infile: %s\n",lineptr);
			exit(156);
		}	
	}
	
	fclose(infile_fd);
//	fprintf(stdout,"MaxPs=%i\n",numPs);
	
	//and start scheduling
	int oldTime=0,Time=0,currP=0;
	//init currentProc to 0
	pproc cp = procs[currP];
	//bool for deciding if currP is 2B blocked in IO
	short bBlockIO=0;
		
	while(!0){
		fprintf(stdout,"At time = %i\n",Time);
		//dont block at time 0, but at any other
		if (Time > 0)
			bBlockIO = 1;
		
			//if RR check if not end of burst (prrempted)
			if ( (sched_meth == RR) && (cp->rung_till > Time) && (cp->state==RUNNING)){
					//advance running counter
					cp->tot_running += (Time - cp->rung_since);
					//decrease remaining burst time
					cp->rung_for -= (Time - cp->rung_since);
					//preempt, ie ...
					bBlockIO = 0;			//do not block in IO
					cp->state = READY;		//and set back to ready
					cp->ready_since = Time;	//remember for waiting time
					fprintf(stdout," - Proc. %i preempted (CPU burst remaining=%i, total CPU remaining %i)\n",
							currP,(cp->rung_till - Time),(cp->C_total - cp->tot_running));
			}
			//if RUNNING and to be blocked
			if (bBlockIO && (cp->state==RUNNING)){
				//advance running counter, after burst
				cp->tot_running += (Time - cp->rung_since);
				//finished? set to TERM
				if ( (cp->tot_running == cp->C_total) && (cp->state!=TERMINATED)){
					cp->state = TERMINATED;
					cp->term_at = Time;
					fprintf(stdout," - process %i terminated\n",currP);
				}else{
					//get next rand int
					int blockfor = udri(cp->IO_max,randfile_fd);
					//and set to sleep, remember stuff for counters
					cp->state = BLOCKED;
					cp->blocked_till = Time+blockfor;
					cp->blocked_since = Time;
					cp->tot_io_time += blockfor;
					cp->rung_for = 0;
					fprintf(stdout," - Proc. %i CPU burst (started at:%i) ended (total CPU remaining %i)\n",
							currP,(cp->rung_since),(cp->C_total - cp->tot_running));
					fprintf(stdout," - Proc. %i I/O started (blocked until %i) (IO burst remaining=%i, total CPU remaining %i)\n",
							currP,(cp->blocked_till),(blockfor),(cp->C_total - cp->tot_running));
				}
			}
		
		//check and set ready procs to ready
		for (i = 0; i < numPs; i++){
			//only if it was blocked and unblocked already or unblocking now
			if ( (procs[i]->state==BLOCKED) && (procs[i]->blocked_till <= Time) ){
				procs[i]->state = READY;
				procs[i]->ready_since = procs[i]->blocked_till;
				fprintf(stdout," - Proc. %i I/O completed (blocked since %i) (total CPU remaining %i)\n",
						i,(procs[i]->blocked_since),(procs[i]->C_total - procs[i]->tot_running));

			}
		}
		
		//get next schedulable task, from sched algo
		currP = sched_next(sched_meth,Time,currP,procs,numPs);
		//fprintf(stdout,"   Next() returned=%i, state=%i\n",currP,procs[currP]->state);
		if (-1 == currP)
			break;			//no more, break
		
		//set currentProc
		cp = procs[currP];
		
		//see if it is ready, might not be, bceause in case of UNI1 might be same
		if (cp->state == READY){
			//count towards ready time
			cp->tot_ready_time += Time - cp->ready_since;

			//rand time for burst
			int burstfor;
			//if was preempter means there is still some time left in CPU burst, take that instead
			if (cp->rung_for>0)
				burstfor = cp->rung_for;
			else			//no: get new rand
				burstfor = udri(cp->B_burst_max,randfile_fd);
			
			//if greater than total time remaining, reduce
			if (burstfor > cp->C_total - cp->tot_running)
				burstfor = cp->C_total - cp->tot_running;
			
			//and set to running, rememb. counters
			cp->state = RUNNING;
			cp->rung_till = Time+burstfor;
			cp->rung_since = Time;
			cp->rung_for = burstfor;
			fprintf(stdout," - process %i CPU burst started (until %i) (CPU burst remaining=%i, total CPU remaining %i)\n",
				currP,(cp->rung_till),(burstfor),(cp->C_total - cp->tot_running));
			

			//save old time
			oldTime = Time;
			//and set time=next event time
			if (sched_meth == RR)			//in case of RR it is min(remaining bursttime,q)
				Time += min(rr_q,burstfor);
			else							//or just burst, because all others dont preempt
				Time += burstfor;
	
		}else{							//if currentP not ready, save time and..
			//just set next event
			oldTime = Time;
			Time = cp->blocked_till;
		}
	
	}//finished (no more procs)
	
	//print stats about procs
	char * algo = malloc(strlen("Uni-programmed 1"));
	strcpy(algo,"Uni-programmed 1");
	
	switch(sched_meth){
		case 1:
			strcpy(algo,"Uni-programmed 2");
			break;
		case 2:
			strcpy(algo,"RR");
			break;
		case 3:
			strcpy(algo,"HPRN");
			break;
	}
	//some space for printing of q
	char q[32] = "";
	if (sched_meth==RR)
		sprintf(q,"With q=%i",rr_q);
	
	printf("-----------------------------------------------------\n");
		
	printf("Simulated %i processes with %s scheduling algorithm. %s\n",numPs,algo,q);
	
	//print out states per process, and sum some counters on the way
	float tot_cpu_busy = 0;
	float tot_turn = 0;
	float tot_wait = 0;
	for (i = 0; i < numPs; i++){
		pproc cp = procs[i];
		printf("--- Process #%i (A=%i, B=%i, C=%i, IO=%i) ---\n",i,cp->A_arriv,cp->B_burst_max,cp->C_total,cp->IO_max);
		printf("    Finished at: \t%i\n",cp->term_at);
		printf("    Turnaround:	 \t%i\n",(cp->term_at - cp->A_arriv));
		printf("    IO Time:	 \t%i\n",(cp->tot_io_time));
		printf("    Wait Time:	 \t%i\n\n",(cp->tot_ready_time));
		tot_cpu_busy += cp->tot_running;
		tot_turn += cp->term_at - cp->A_arriv;
		tot_wait += cp->tot_ready_time;
	}
	
	//and general stats
	float cpu_util = (tot_cpu_busy/Time * 100);
	float io_util =  (100 - cpu_util);
	printf("--- General stats ---\n");
	printf("    Finish Time:	\t%i\n", Time);
	printf("    CPU Util.:		\t%f\n", cpu_util);
	printf("    IO Util.:		\t%f\n", io_util);
	printf("    Thru-put:		\t%f\n", (((float)Time/(float)numPs)/100) );
	printf("    Avg. turn Time:	\t%f\n", (tot_turn/numPs) );
	printf("    Avg. wait Time:	\t%f\n", (tot_wait/numPs) );
	
}

/*
just prints out how to use this sceduler simu
reads its input from file usage.txt
*/
void usage(){
	char opentype = 'r';
	int bufsize = 64;
	char lineptr[bufsize];
		
	FILE * fd = fopen("usage.txt", &opentype);
	if (fd > 0)				//print out each line
		while ( 0 < fgets((char *)&lineptr,bufsize,fd) )
			printf("%s",lineptr);
			
	fclose(fd);
}
