/* sched_algo.c
 * simple scheduler Algorithm implementations
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <math.h>

#include "proc.h"


int sched_next(int algo, int t, int p, pproc ps[], int max){
	//return value
	int ret;
	int i;
	//bools for check if all dead, and not all Blocked (bNotBlocked)
	short bActive,bNBlocked;
	
	//fprintf(stdout,"   called s_next with algo=%i,t=%i,p=%i,max=%i\n",algo,t,p,max);

	//select method
	switch(algo){
		
		case UNI1:
			//just return same if not TERM
			if (ps[p]->state != TERMINATED)
				return p;
			//go to next
			ret=p+1;
			//if last, signal termination
			if (ret == max)
				return -1;
			
			return ret;
			
		case UNI2:
		case RR:
			
			bActive=bNBlocked=0;
			//min. saved unblock time
			int minevt = 32000;
			//and which to return if all are blocked
			int ret_if_allblocked = -1;
			
			//at t=0 return input
			if (!t) return p;
			
			//check if any are active, and all runnable are blocked
			for (i = 0; i < max; i++){
				if (ps[i]->state != TERMINATED) bActive=1;
				if (ps[i]->state == READY && ps[i]->A_arriv<=t) bNBlocked=1;
				//and remember min-unBlock just in case
				if ( (ps[i]->state == BLOCKED) 		&& 
					 (ps[i]->blocked_till < minevt) && 
					 (ps[i]->A_arriv<=t)
				   ){
					minevt = ps[i]->blocked_till;
					ret_if_allblocked = i;
				}
			}
			//all terminated
			if (!bActive) return -1;
			//all blocked, return smallest unblock time
			if (!bNBlocked) return ret_if_allblocked;
			//just one proc and not blocked|TERM, return it
			if (max==1) return p;
			
			//now we can safely assume that there is a schedulable proc - find it!
			ret = ((p+1) % max);
			//run over whole array until suitable proc is found, that has already started, and is ready
			while ( (ps[ret]->state!=READY) || (ps[ret]->A_arriv>t) )
				ret = ((ret+1) % max);

			return ret;
						
		case HPRN:
		
		 	bActive=0;
			float high = 0;				//highest current R
			ret = p;					//and its thread idx

			//loop over all threads and find the one with the highest Penal. ratio
			for (i = 0; i < max; i++){
				float r;
				if (ps[i]->tot_running == 0)
					r = INFINITY;				//if div by 0, set to INF, does not seem to happen autom.
				else
					r = ( (t-ps[i]->A_arriv) / ps[i]->tot_running);
				
				//fprintf(stdout,"   PR(%i)=%f %s\n",i,r, ((ps[i]->state==TERMINATED)?"(TERMINATED)":"") );
				//if higher remember
				if ( ( r > high) && (ps[i]->state != TERMINATED) ){
					high = r;
					ret = i;
				}
				//if there is at least 1 !TERM, set active
				if (ps[i]->state!=TERMINATED)
					bActive = 1;
			}
			if (bActive)
				return ret;
			//not active, so return -1
			return -1;
	}
	
	return -1;

}
