#include "tthreads.h"
#include "stdio.h"
#include "mutex.h"

#define NUM_WORKERS 4
#define NUM_CELLS	10 
#define WORKER_MIN	0 
#define WORKER_MAX	100 

//mutex for access to cells
Mutex m;
//active workers count, for terminating the consumer
short active;

//the cells
typedef struct {
  int *cells;
  int *filledCells;
} TCells;

//the workers
typedef struct {
  int workerNum;
  //startVal that this worker is at (WORKER_MIN..WORKER_MAX)
  int workerValue;
  //internal pointer to the cells
  TCells *workCells;
} TWork;

/* Function for workers, gets started NUM_WORKERS times
 * worker#0 is consumer
*/
void worker(TWork *w)
{
  int workerNum = w->workerNum;
  int workerValue = WORKER_MIN;
  //global counter in workcells how many are filled
  int *filledCells = w->workCells->filledCells;
  //the cells themselves
  int *cells = w->workCells->cells;
  
  if (workerNum) {		/* start next worker (num# - 1) until 0 reached*/
    //init struct
    TWork *nextWork = malloc(sizeof(TWork));
    //set its number
    nextWork->workerNum = workerNum - 1;
    nextWork->workCells = w->workCells;
    //and schedule
    initAndScheduleThread((void*)(void*)worker, nextWork);
    yield();			/* vollunteer for preemption */
  }

  if (!workerNum) {		//consumer
    yield();
    while (active){
	    printf("+++++++++ Consumer starting. %d filled cells ++++++++\n",(*filledCells));
    	//enter critical section
    	mutexLock(&m);
    	printf("Consumer consumed ");
    	//and consume all that are there
		while( (*filledCells) >= 0){
			printf("cell[%d]=%d; ", (*filledCells),cells[(*filledCells)]);
			//add its number to the sum
			workerValue += cells[(*filledCells)];
			//and dec. number of filled cells
			(*filledCells)--;
		}
      	printf("\n");
		fflush(stdout);
	    mutexUnlock(&m);		//unlock cells - leave crit. section
      	yield();				//and preempt
    }
    
  } else {
    while ( workerValue < WORKER_MAX ){		//if not reached WORKER_MAX
      mutexLock(&m);						//enter crit.section
      printf("----------- Producer %d Running, startVal=%d ---------\n", workerNum,workerValue);
      printf("Producer %d filled ",workerNum);
      //as long as there are free cells and we havent reached WORKER_MAX
      //only count to NUM_CELLS-2 coz filled_cells is inc. while filling
      //off-by-one error otherwise
      while (((*filledCells) < (NUM_CELLS-1)) && (workerValue < WORKER_MAX)){
		 cells[++(*filledCells)] = (workerValue++);			//fill next cell
      	 printf("cell[%d]=%d; ",(*filledCells),workerValue);
	  }
      printf("\n");
      fflush(stdout);
	  yield();				//and preempt
      mutexUnlock(&m);		//leave crit. section
      
    }
  }
  //#num terminated - dec. active Producer count
  if (workerNum){
  	active--;
	printf("Producer %d terminated\n", workerNum);
  }else		//consumer terminated
  	printf("Consumer terminated, Sum=%d\n",workerValue);
  free(w);
  return;
}

int main()
{
  //init mutex with one thread allowed
  mutexInit(&m, 1);
  int numThreads = NUM_WORKERS+1;
  makeThreads(numThreads);
  
  active = numThreads-1;
  
  TWork *work = malloc(sizeof(TWork));
  work->workerNum = numThreads-1;
  work->workCells = malloc(sizeof(TCells));
  work->workCells->cells= malloc(NUM_CELLS * sizeof(int));
  work->workCells->filledCells = malloc(sizeof(int));
  *(work->workCells->filledCells) = 0;
  
  initAndScheduleThread((void*)(void*)worker, work);
  
  startThreads();		/* start working! */
  //ended
  printf("Threads Terminated\n");
}
