#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/resource.h>

#include "myAllocator.h"

//some defs
#define TEST_ALLOC 	8
#define SMALL_ALLOC 256
#define SMALL_INC 	128
#define BIG_ALLOC 	SMALL_ALLOC*100
#define BIG_INC 	SMALL_INC*100
#define MEAS_TIMES	10000

//names
#define FIRST		0
#define NEXT		1
#define BEST		2


//diffs 2 timevals
double diffTimeval(struct timeval *t1, struct timeval *t2) {
  double d = (t1->tv_sec - t2->tv_sec) + (1.0e-6 * (t1->tv_usec - t2->tv_usec));
  return d;
}

//returns current time in given timeval struct
void getutime(struct timeval *t) {
  struct rusage usage;
  getrusage(RUSAGE_SELF, &usage);
  *t = usage.ru_utime;
}

//main prog
int main() {

	//just var to remember which algo to use
  int toDo;
  //make array of the 3 different funcs to use, and initialize
  findFitFunc fitFuncArr[BEST+1] = {	&findFirstFit,
  										&findNextFit,
  										&findBestFit};
  
  //and loop over 4 algos
  for(toDo = FIRST; toDo <= BEST; toDo++){ 
    
    //cahr for name of algo
	char* algo;
	
    switch(toDo){
    	case FIRST:
    		algo = "firstFit";
    		break;
    	case NEXT:
    		algo = "nextFit";
    		break;
    	case BEST:
    		algo = "bestFit";
    		break;
    }
    
    printf("\n---- Using %s ----\n", algo);
    
    void *p1, *p2, *p3, *p4;
	arenaCheck();
    
    //start by allocating some space
    printf("Allocating %db...\n", SMALL_ALLOC);
    p1 = AllocRegionPtr(SMALL_ALLOC,fitFuncArr[toDo]);
    arenaCheck();
    
    printf("Allocating %db...\n", BIG_ALLOC);
    p2 = AllocRegionPtr(BIG_ALLOC,fitFuncArr[toDo]);
    arenaCheck();
    
    printf("Allocating %db...\n", SMALL_ALLOC);
    p3 = AllocRegionPtr(SMALL_ALLOC,fitFuncArr[toDo]);
    arenaCheck();

    printf("Allocating %db...\n", BIG_ALLOC);
    p4 = AllocRegionPtr(BIG_ALLOC,fitFuncArr[toDo]);
    printf("p1=0x%x, p2=0x%x, p3=0x%x, p4=0x%x\n", p1, p2, p3, p4);
    arenaCheck();
    
    //now try resizing a little bit
    printf("Resizing p3=0x%x (%db+%db=%db) ...\n", p3, SMALL_ALLOC, SMALL_INC, SMALL_ALLOC+SMALL_INC);
    p3 = resizeRegionPtr(p3,SMALL_ALLOC+SMALL_INC,fitFuncArr[toDo]);
    printf("p1=0x%x, p2=0x%x, p3=0x%x, p4=0x%x\n", p1, p2, p3, p4);
    arenaCheck();
    
    printf("Resizing p1=0x%x (%db+%db=%db) ...\n", p1, SMALL_ALLOC, BIG_INC, BIG_INC+SMALL_ALLOC);
    p1 = resizeRegionPtr(p1,SMALL_ALLOC+BIG_INC,fitFuncArr[toDo]);
    printf("p1=0x%x, p2=0x%x, p3=0x%x, p4=0x%x\n", p1, p2, p3, p4);
    arenaCheck();
    
    printf("Resizing p2=0x%x (%db+%db=%db) ...\n", p2, BIG_ALLOC, BIG_INC, BIG_INC+BIG_ALLOC);
    p2 = resizeRegionPtr(p2,BIG_ALLOC+BIG_INC,fitFuncArr[toDo]);
    printf("p1=0x%x, p2=0x%x, p3=0x%x, p4=0x%x\n", p1, p2, p3, p4);
    arenaCheck();

    printf("Resizing p4=0x%x (%db+%db=%db) ...\n", p4, BIG_ALLOC, SMALL_INC, SMALL_INC+BIG_ALLOC);
    p4 = resizeRegionPtr(p4,BIG_ALLOC+SMALL_INC,fitFuncArr[toDo]);
    printf("p1=0x%x, p2=0x%x, p3=0x%x, p4=0x%x\n", p1, p2, p3, p4);
    arenaCheck();
    
    //and free stuff again
    printf("Freeing p2=0x%x...\n", p2);
    freeRegion(p2);
    arenaCheck();

    printf("Freeing p3=0x%x...\n", p3);
    freeRegion(p3);
    arenaCheck();

    printf("Freeing p1=0x%x...\n", p1);
    freeRegion(p1);
    arenaCheck();

    printf("Freeing p4=0x%x...\n", p4);
    freeRegion(p4);
    arenaCheck();
    
    {				// measure time for MEAS_TIMES mallocs
      struct timeval t1, t2;
      int i;
      getutime(&t1);
	  for(i = 0; i < MEAS_TIMES; i++)
		if ( AllocRegionPtr(TEST_ALLOC, fitFuncArr[toDo]) == 0 ) 
	  		break;
	  		
      getutime(&t2);
      printf("%dx %s(4) required %f seconds\n ---- end of %s ---- \n", i, algo, diffTimeval(&t2, &t1),algo);
    }
    
    if (toDo<BEST)
    	clearArena();
  }//end while
  
  return 0;
}//end main
